from typing import Collection

from buz.command import Command, CommandHandler


class MoreThanOneCommandHandlerRelatedException(Exception):
    def __init__(self, command: Command, command_handlers: Collection[CommandHandler]):
        self.command = command
        self.command_handlers = command_handlers
        super().__init__(f"There is more than one handler registered for {command.fqn()}.")



from buz.command.sync.more_than_one_command_handler_related_exception import MoreThanOneCommandHandlerRelatedException
from buz.command.sync.sync_command_bus import SyncCommandBus

__all__ = [
    "SyncCommandBus",
    "MoreThanOneCommandHandlerRelatedException",
]



from buz.query.sync.more_than_one_query_handler_related_exception import MoreThanOneQueryHandlerRelatedException
from buz.query.sync.sync_query_bus import SyncQueryBus

__all__ = [
    "SyncQueryBus",
    "MoreThanOneQueryHandlerRelatedException",
]



from typing import Collection

from buz.query import Query, QueryHandler


class MoreThanOneQueryHandlerRelatedException(Exception):
    def __init__(self, query: Query, query_handlers: Collection[QueryHandler]):
        self.query = query
        self.query_handlers = query_handlers
        super().__init__(f"There is more than one handler registered for {query.fqn()}.")



from logging import Logger

from buz.event import EventBus
from buz.event.kombu import EventNotPublishedException
from buz.event.transactional_outbox import (
    OutboxRepository,
    OutboxCriteria,
    OutboxRecordToEventTranslator,
    OutboxRecordStreamFinder,
)
from buz.event.transactional_outbox.outbox_criteria import OutboxSortingCriteria
from buz.locator import MessageFqnNotFoundException


class TransactionalOutboxWorker:
    def __init__(
        self,
        outbox_repository: OutboxRepository,
        outbox_record_stream_finder: OutboxRecordStreamFinder,
        outbox_record_to_event_translator: OutboxRecordToEventTranslator,
        event_bus: EventBus,
        logger: Logger,
    ):
        self.__outbox_repository = outbox_repository
        self.__outbox_record_stream_finder = outbox_record_stream_finder
        self.__outbox_record_to_event_translator = outbox_record_to_event_translator
        self.__event_bus = event_bus
        self.__logger = logger

    def start(self) -> None:
        criteria = OutboxCriteria(delivered_at=None, order_by=OutboxSortingCriteria.CREATED_AT)
        for outbox_record in self.__outbox_record_stream_finder.find(criteria):
            try:
                event = self.__outbox_record_to_event_translator.translate(outbox_record)

                self.__event_bus.publish(event)
                outbox_record.mark_as_delivered()

            except (EventNotPublishedException, MessageFqnNotFoundException, Exception) as e:
                self.__logger.exception(e)
                outbox_record.mark_delivery_error()

            self.__outbox_repository.save(outbox_record)



from datetime import datetime
from typing import Dict, Optional, Any, ClassVar

from dataclasses import dataclass


@dataclass
class OutboxRecord:  # type: ignore[misc]
    DATE_TIME_FORMAT: ClassVar[str] = "%Y-%m-%d %H:%M:%S.%f"

    event_id: str
    event_fqn: str
    event_payload: Dict[str, Any]  # type: ignore[misc]
    created_at: datetime
    delivered_at: Optional[datetime] = None
    delivery_errors: int = 0

    def parsed_created_at(self) -> str:
        return self.created_at.strftime(self.DATE_TIME_FORMAT)

    def mark_delivery_error(self) -> None:
        self.delivered_at = None
        self.delivery_errors += 1

    def mark_as_delivered(self) -> None:
        self.delivered_at = datetime.now()



from buz.command.command import Command
from buz.command.command_handler import CommandHandler
from buz.command.base_command_handler import BaseCommandHandler
from buz.command.command_bus import CommandBus

__all__ = [
    "Command",
    "CommandHandler",
    "BaseCommandHandler",
    "CommandBus",
]



from typing import Collection

from buz.event import Event, EventBus
from buz.event.transactional_outbox.event_to_outbox_record_translator import EventToOutboxRecordTranslator
from buz.event.transactional_outbox.outbox_repository import OutboxRepository


class TransactionalOutboxEventBus(EventBus):
    def __init__(
        self,
        outbox_repository: OutboxRepository,
        event_to_outbox_record_translator: EventToOutboxRecordTranslator,
    ):
        self.__outbox_repository = outbox_repository
        self.__event_to_outbox_record_translator = event_to_outbox_record_translator

    def publish(self, event: Event) -> None:
        outbox_record = self.__event_to_outbox_record_translator.translate(event)
        self.__outbox_repository.save(outbox_record)

    def bulk_publish(self, events: Collection[Event]) -> None:
        for event in events:
            self.publish(event)



from collections import defaultdict
from typing import List, DefaultDict, Generic, TypeVar, Type

from pypendency.container import AbstractContainer

from buz import Handler
from buz import Message
from buz.locator import Locator, HandlerFqnNotFoundException, MessageFqnNotFoundException
from buz.locator.pypendency import HandlerAlreadyRegisteredException
from buz.locator.pypendency import HandlerNotFoundException
from buz.locator.pypendency import HandlerNotRegisteredException

K = TypeVar("K", bound=Message)
V = TypeVar("V", bound=Handler)
MessageFqn = str


class ContainerLocator(Locator, Generic[K, V]):
    CHECK_MODE_REGISTER_TIME = "register"
    CHECK_MODE_GET_TIME = "get"

    def __init__(self, container: AbstractContainer, check_mode: str = CHECK_MODE_REGISTER_TIME) -> None:
        self.__container = container
        self.__check_mode = check_mode
        self.__mapping: DefaultDict[MessageFqn, List[V]] = defaultdict(list)
        self.__handler_ids: List[str] = []
        self.__handlers_resolved = False

    def register(self, handler_id: str) -> None:
        self.__guard_handler_already_registered(handler_id)
        if self.__check_mode == self.CHECK_MODE_REGISTER_TIME:
            self.__guard_handler_not_found(handler_id)
        self.__handler_ids.append(handler_id)
        self.__handlers_resolved = False

    def __guard_handler_already_registered(self, handler_id: str) -> None:
        if handler_id in self.__handler_ids:
            raise HandlerAlreadyRegisteredException(handler_id)

    def __guard_handler_not_found(self, handler_id: str) -> None:
        if not self.__container.has(handler_id):
            raise HandlerNotFoundException(handler_id)

    def unregister(self, handler_id: str) -> None:
        self.__guard_handler_not_registered(handler_id)
        self.__handler_ids.remove(handler_id)
        self.__handlers_resolved = False

    def __guard_handler_not_registered(self, handler_id: str) -> None:
        if handler_id not in self.__handler_ids:
            raise HandlerNotRegisteredException(handler_id)

    def get(self, message: K) -> List[V]:
        self.__ensure_handlers_resolved()
        return self.__mapping.get(message.fqn(), [])

    def __ensure_handlers_resolved(self) -> None:
        if not self.__handlers_resolved:
            self.__resolve_handlers()

    def __resolve_handlers(self) -> None:
        self.__mapping = defaultdict(list)
        for handler_id in self.__handler_ids:
            if self.__check_mode == self.CHECK_MODE_GET_TIME:
                self.__guard_handler_not_found(handler_id)
            handler: V = self.__container.get(handler_id)
            message_fqn = handler.handles().fqn()
            self.__mapping[message_fqn].append(handler)
        self.__handlers_resolved = True

    def get_handler_by_fqn(self, handler_fqn: str) -> V:
        self.__ensure_handlers_resolved()
        for message_handlers in self.__mapping.values():
            for handler in message_handlers:
                if handler_fqn == handler.fqn():
                    return handler
        raise HandlerFqnNotFoundException(handler_fqn)

    def get_message_klass_by_fqn(self, message_fqn: str) -> Type[K]:
        self.__ensure_handlers_resolved()
        try:
            handler = self.__mapping.get(message_fqn, [])[0]
            return handler.handles()
        except (IndexError, TypeError, HandlerFqnNotFoundException):
            raise MessageFqnNotFoundException(message_fqn)
